
//Author Stavros Martini TH20367

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <pthread.h>
#include <sched.h>
#include <string.h>
#include <stdbool.h>
#include <time.h>
#include <math.h>

#define FLIGHTS_NUM 10  // Number of flights
#define N 15          // Number of child processes (agents)
#define SEATS 3     // Number of seats per flight
#define FRAMES  2  // for the plotting resolution
FILE *gnuplotPIPE = NULL;  // pipe to give commands to gnuplot 

typedef struct {
    char flight_route[30]; // Flight route string, e.g., "ATH -> HER"
    bool seats[SEATS];     // Array to hold reservation status of each seat
} Flight;

typedef struct {
    Flight flights[FLIGHTS_NUM];
    int last_reserved_seat; // Index of the last reserved seat
    int last_visited_agent; // ID of the agent who made the last reservation
    int last_flight_index;  // Index of the last flight with a reservation
    int reserve_count;      // Total number of reservations made
    pthread_mutex_t lock;
    pthread_cond_t cond;
} SharedMemory;

void init_plot() {
    // Initialize the pipe to GNUplot
    gnuplotPIPE = popen("gnuplot -persistent", "w");
    if (gnuplotPIPE == NULL) {
        fprintf(stderr, "Failed to open pipe to GNUplot\n");
        exit(EXIT_FAILURE);
    }

    // Setup the plot environment
    fprintf(gnuplotPIPE, "set terminal qt size 600,400\n");  // Set the output terminal
    fprintf(gnuplotPIPE, "set grid\n");                       // Enable grid
    fprintf(gnuplotPIPE, "set title 'Real-time Data Plot'\n");  // Set title
    fprintf(gnuplotPIPE, "set xlabel 'Time secs'\n");
    fprintf(gnuplotPIPE, "set ylabel 'Reserved-Seats'\n");
    fprintf(gnuplotPIPE, "set key off\n");
}

void replot(const char *dataFileName) {
    // Check if the pipe is initialized
    if (gnuplotPIPE == NULL) {
        fprintf(stderr, "GNUplot not initialized.\n");
        return;
    }

    // Command to plot data from a file
    fprintf(gnuplotPIPE, "plot '%s' using 1:2 with linespoints linewidth 2 lc rgb \"black\"\n", dataFileName);
    fflush(gnuplotPIPE);  // Flush the pipe to ensure GNUplot processes commands
}
void fit_and_replot(const char *dataFileName, double total_reservations) {
    if (gnuplotPIPE == NULL) {
        fprintf(stderr, "GNUplot not initialized.\n");
        return;
    }

    fprintf(gnuplotPIPE, "set fit quiet\n");  // Suppress fit notifications
    fprintf(gnuplotPIPE, "f(x) = a*x + b\n");  // Define the fitting function
    fprintf(gnuplotPIPE, "fit f(x) '%s' using 1:2 via a, b\n", dataFileName);  // Perform the fit
    // Tell Gnuplot to calculate x_value based on the total_reservations and fitted parameters
    fprintf(gnuplotPIPE, "x_value = (%.17g - b) / a\n", total_reservations);  // Calculate the x_value
    fprintf(gnuplotPIPE, "set label 1 sprintf('End time aprox: %%.4f sec', x_value) at graph 0.1, 0.9 front\n");  // Set the label with x_value
    fprintf(gnuplotPIPE, "set key box\n");  // Box around the legend
    // Plot the data
    fprintf(gnuplotPIPE, "set title 'Real-time Data Plot'\n");
    fprintf(gnuplotPIPE, "set yrange [0:*]\n");  // Ensure only positive y-values are displayed
    fprintf(gnuplotPIPE, "plot '%s' using 1:2 with linespoints title 'Data Points', \\\n", dataFileName);
    fprintf(gnuplotPIPE, "f(x) with lines dt 2 title 'Fitted line'\n");

    fflush(gnuplotPIPE);  // Ensure commands are processed by Gnuplot
}





void init_shared_memory(SharedMemory *shm) {
    pthread_mutexattr_t attr;
    pthread_condattr_t cond_attr;
    //mutex init
    pthread_mutexattr_init(&attr);
    pthread_mutexattr_setpshared(&attr, PTHREAD_PROCESS_SHARED);
    pthread_mutex_init(&shm->lock, &attr);
    pthread_mutexattr_destroy(&attr);
    //condition variable
    pthread_condattr_init(&cond_attr);
    pthread_condattr_setpshared(&cond_attr, PTHREAD_PROCESS_SHARED);
    pthread_cond_init(&shm->cond, &cond_attr);
    pthread_condattr_destroy(&cond_attr);
    //for the sake of the demo
    const char* flight_routes[FLIGHTS_NUM] = {"ATH -> HER", "NYC -> LAX", "LON -> PAR", "BER -> MUC", "SYD -> MEL",
                                              "MIA -> SEA", "DFW -> SFO", "LAS -> BOS", "CDG -> JFK", "HND -> SFO"};
    for (int i = 0; i < FLIGHTS_NUM; i++) {
        strcpy(shm->flights[i].flight_route, flight_routes[i]); //copy routes
        memset(shm->flights[i].seats, false, sizeof(shm->flights[i].seats));//zeroout
    }
    shm->reserve_count = 0;  // Initialize reservation count , and rest flags
    shm->last_flight_index=-1;
    shm->last_reserved_seat=-1;
    shm->last_visited_agent=-1;
}

//clean shared mem
void cleanup(SharedMemory *shm, int shm_id) {
    pthread_mutex_destroy(&shm->lock);//clean mutex
    pthread_cond_destroy(&shm->cond); //clean cond var
    shmdt(shm);  
    shmctl(shm_id, IPC_RMID, NULL);
}

//booking 
void book_seat(int agent_id, SharedMemory *shm, int flight_index, int seat_num) {
    pthread_mutex_lock(&shm->lock); //lock
    if (!shm->flights[flight_index].seats[seat_num]) {     //book
        shm->flights[flight_index].seats[seat_num] = true;
        shm->last_reserved_seat = seat_num;
        shm->last_visited_agent = agent_id;
        shm->last_flight_index = flight_index;
        shm->reserve_count++;
        pthread_cond_signal(&shm->cond);
    }
    pthread_mutex_unlock(&shm->lock); //unlock
}

void child_work(int pipe_read_end, int agent_id, SharedMemory *shm) {
    int data[2];  // To receive flight index and seat number
    srand(time(NULL)+getpid());//random seed generator
    while (read(pipe_read_end, &data, sizeof(data)) > 0) {  //wait unitl write happens in pipe

        u_int32_t delay =(u_int32_t) ( (rand()+rand())*rand()*(!rand() + rand()/((rand()^rand())+1)  ) ); //rand uint function
        usleep( delay%1500001 +  600000);  // Simulate variable delay for processing [0.6 sec <= delay <= 2.1 sec]
        book_seat(agent_id, shm, data[0], data[1]); //bokk the flight seat given 
    }
    close(pipe_read_end);
    exit(0);
}
int get_assigned_cpu(pid_t pid) {
    cpu_set_t cpuset;
    CPU_ZERO(&cpuset);
    if (sched_getaffinity(pid, sizeof(cpu_set_t), &cpuset) < 0) {
        perror("sched_getaffinity failed");
        exit(EXIT_FAILURE);
    }
    for (int i = 0; i < CPU_SETSIZE; i++) {
        if (CPU_ISSET(i, &cpuset)) {
            return i; // Return the first core found
        }
    }
    return -1; // If no core found, but should not happen
}

void create_child_processes(int pipe_fd[N][2], pid_t pids[N], SharedMemory *shm) {
    printf("Parent Process (PID:%4d)\n", getpid());
    for (int i = 0; i < N; i++) {
        // Create two pipes for each child, one for parent-child sync
        int sync_pipe[2];
        if (pipe(pipe_fd[i]) < 0 || pipe(sync_pipe) < 0) {
            perror("Pipe creation failed");
            exit(1);
        }

        pids[i] = fork();
        if (pids[i] < 0) {
            perror("Fork failed");
            exit(1);
        } else if (pids[i] == 0) { // Child process
            close(pipe_fd[i][1]);  // Child closes the write end of the main pipe
            close(sync_pipe[0]);  // Child closes the read end of the sync pipe

            cpu_set_t cpuset;
            CPU_ZERO(&cpuset);  // Clear all entries in the CPU set
            CPU_SET(i % sysconf(_SC_NPROCESSORS_ONLN), &cpuset);  // Set the CPU bit for this child process

            if (sched_setaffinity(0, sizeof(cpu_set_t), &cpuset) < 0) {
                perror("sched_setaffinity failed");
                exit(EXIT_FAILURE);
            }

            // Signal to parent that CPU affinity is set by writing to the sync pipe
            char ready_signal = '1';
            if (write(sync_pipe[1], &ready_signal, sizeof(char)) < 0) { //notify parent that child got assinged to a core
                perror("Write to sync pipe failed");
                exit(EXIT_FAILURE);
            }
            close(sync_pipe[1]);  // Close the write end after sending signal

            child_work(pipe_fd[i][0], i, shm);
            exit(0);  // Ensure child exits after finishing work
        } else { // Parent process
            close(pipe_fd[i][0]);  // Parent closes the read end of the main pipe
            close(sync_pipe[1]);  // Parent closes the write end of the sync pipe

            // Wait for child to signal readiness
            char read_buffer;
            if (read(sync_pipe[0], &read_buffer, sizeof(char)) < 0) {  // father wait's until write happens in sync_pipe (child gets assinged to core)
                perror("Read from sync pipe failed");
                exit(EXIT_FAILURE); // did not get assigned
            }
            close(sync_pipe[0]);  // Close the read end after receiving signal

            printf("(AGENT:%2d) Generated Child Process (PID:%4d) assigned to core(No:%d)\n", i, pids[i], get_assigned_cpu(pids[i]));
            fflush(stdout);
        }
    }
}


//same as up but more simple (less random) and old version maybe does not work (kept here more for understanding of simples princibles)
//Linear distrubution to random agent
/*void parent_process(pid_t pids[N],int pipe_fd[N][2], SharedMemory *shm, struct timespec start_time) {
    FILE *fp = fopen("reservations.dat", "a");
    struct timespec now;
    if (fp == NULL) {
        perror("Failed to open reservations file");
        exit(EXIT_FAILURE);
    }

    // Start by writing all seat assignments to the pipes
    for (int f = 0; f < FLIGHTS_NUM; f++) {
        for (int s = 0; s < SEATS; s++) {
            int child_id = rand() % N;  // Randomly select an agent for each seat
            int data[2] = {f, s};
            write(pipe_fd[child_id][1], &data, sizeof(data));

            // Lock and wait for confirmation of the booking
            pthread_mutex_lock(&shm->lock);
            // Continuously check the condition while handling spurious wake-ups
            while (shm->last_reserved_seat != s || shm->last_flight_index != f) {
                pthread_cond_wait(&shm->cond, &shm->lock);
            }

            // Calculate the elapsed time from the start of the program to the booking
            clock_gettime(CLOCK_MONOTONIC, &now);
            double elapsed_time = (now.tv_sec - start_time.tv_sec) + (now.tv_nsec - start_time.tv_nsec) / 1e9;
            // Print booking information to the console
            printf("time: %3.4f sec | [Agent:%2d (PID:%d)] booked | (seat %2d [%s]) | Total bookings: %3d\n",
                   elapsed_time, shm->last_visited_agent,pids[shm->last_visited_agent], shm->last_reserved_seat,
                   shm->flights[shm->last_flight_index].flight_route, shm->reserve_count);
            fflush(stdout);

            // Write the elapsed time and the total reservations to the file
            fprintf(fp,"%3.8f %d\n",elapsed_time, shm->reserve_count);
            fflush(fp);
            if(shm->reserve_count % 3 ==0){replot("reservations.dat");}

            // Unlock after finishing work with shared variables
            pthread_mutex_unlock(&shm->lock);
        }
    }

    if (fclose(fp) != 0) {
        perror("Failed to close reservations file");
        exit(EXIT_FAILURE);
    }

    // Close the write ends of pipes to signal to children that no more data will be sent
    for (int i = 0; i < N; i++) {
        close(pipe_fd[i][1]);
    }

    printf("\n\n=-=-=-=-=-=-= ALL FLIGHT HAVE ALL THEIR SEATS BOOKED =-=-=-=-=-=-=\n\n");
    fflush(stdout);
   
    for (int i=0; i < N; i++){
        int status;
        waitpid(pids[i], &status ,0);
        if (WIFEXITED(status)) {
                printf("(AGENT:%2d) Child Process (PID:%4d) Exited with %d\n",i,pids[i],status);
            } else {
                printf("AGENT:%2d) Child Process (PID:%4d) Terminated abnormaly with %d\n",i,pids[i],status);
            }
    }
}
*/

//Random distrubution to random agent
void parent_process(pid_t pids[N] ,int pipe_fd[N][2], SharedMemory *shm, struct timespec start_time) {
    FILE *fp = fopen("reservations.dat", "a");
    struct timespec now;
    if (fp == NULL) {
        perror("Failed to open reservations file");
        exit(EXIT_FAILURE);
    }
    srand(time(NULL)+getpid()); //rand generator

    // Start by writing all seat assignments to the pipes
    for (int f = 0; f < FLIGHTS_NUM; f++) {
        for (int s = 0; s < SEATS; s++) {
            int flight_index, seat_num;

            // Keep generating random flight and seat pairs until finding a valid one
            do {
                flight_index = rand() % FLIGHTS_NUM;  // Randomly select a flight
                seat_num = rand() % SEATS;  // Randomly select a seat
            } while (shm->flights[flight_index].seats[seat_num]);  // Check if the seat is already reserved

            int agent_id = rand() % N;  // Randomly select an agent

            int data[2] = {flight_index, seat_num};
            write(pipe_fd[agent_id][1], &data, sizeof(data));

            // Lock and wait for confirmation of the booking
            pthread_mutex_lock(&shm->lock);
            // Continuously check the condition while handling spurious wake-ups
            while (shm->last_reserved_seat != seat_num || shm->last_flight_index != flight_index) {
                pthread_cond_wait(&shm->cond, &shm->lock);
            }
            // create the string command to give to system()
            char command[300];
            char mess[150];
            snprintf(mess, sizeof(mess), "Agent %d (pid:%d) reserved the seat %d in flight %s", shm->last_visited_agent, pids[shm->last_visited_agent], shm->last_reserved_seat, shm->flights[shm->last_flight_index].flight_route);
            snprintf(command, sizeof(command), "./my_popup.sh /tmp/call_count \"%s\"", mess);
            system(command);


            // Calculate the elapsed time from the start of the program to the booking
            clock_gettime(CLOCK_MONOTONIC, &now);
            double elapsed_time = (now.tv_sec - start_time.tv_sec) + (now.tv_nsec - start_time.tv_nsec) / 1e9;
            // Print booking information to the console
            printf("time: %3.4f sec | [Agent:%2d (PID:%4d)] booked | (seat %2d [%s]) | Total bookings: %3d\n",
                   elapsed_time, shm->last_visited_agent,pids[shm->last_visited_agent], shm->last_reserved_seat,
                   shm->flights[shm->last_flight_index].flight_route, shm->reserve_count);
            fflush(stdout);

            // Write the elapsed time and the total reservations to the file
            fprintf(fp,"%.8f %d\n",elapsed_time, shm->reserve_count); 
            fflush(fp);
            if(shm->reserve_count % FRAMES == 0){
                fit_and_replot("reservations.dat",FLIGHTS_NUM*SEATS);
            }else if((shm->reserve_count %  FRAMES != 0 )&&(shm->reserve_count == FLIGHTS_NUM*SEATS)){ //last plot
                fit_and_replot("reservations.dat",FLIGHTS_NUM*SEATS);

            }

            // Unlock after finishing work with shared variables
            pthread_mutex_unlock(&shm->lock);  
        }
    }
    //close .dat file
    if (fclose(fp) != 0) {
        perror("Failed to close reservations file");
        exit(EXIT_FAILURE);
    }

    // Close the write ends of pipes EOF to signal to children that no more data will be sent
    for (int i = 0; i < N; i++) {
        close(pipe_fd[i][1]);
    }
    printf("\n\n=-=-=-=-=-=-= ALL FLIGHT HAVE ALL THEIR SEATS BOOKED =-=-=-=-=-=-=\n\n");
    fflush(stdout);
    system("pkill zenity");//terminate the pop up window
   
   //in theory all the child processes(small delays) here have exited so no zombie , but for safety parent waits and captures the exit staus of each child 
    for (int i=0; i < N; i++){
        int status;
        waitpid(pids[i], &status ,0);
        if (WIFEXITED(status)) {
                printf("(AGENT:%2d) Child Process (PID:%4d) Exited with %d\n",i,pids[i],status);
            } else {
                printf("AGENT:%2d) Child Process (PID:%4d) Terminated abnormaly with %d\n",i,pids[i],status);
            }
    }
    
    
}

int main() {
    //shared memory creation
    int shm_id = shmget(IPC_PRIVATE, sizeof(SharedMemory), IPC_CREAT | 0666);
    if (shm_id == -1) {
        perror("shmget failed");
        exit(EXIT_FAILURE);
    }
    //empty the .dat file and write the point 0.00 0
    system("echo '0.000 0' > reservations.dat");//zero point
    //initialize the plot
    init_plot();
    SharedMemory *shm = (SharedMemory *)shmat(shm_id, NULL, 0); //get/request shared memory
    struct timespec start;
    int pipe_fd[N][2]; //pipes
    pid_t pids[N];     //proccess id's

    init_shared_memory(shm);  //zero out shared memory & init mutex , cond var and various flags
    create_child_processes(pipe_fd, pids, shm);  // create child processes 
    //here all child processes are waiting for bytes to be writen in each of their pipe
    clock_gettime(CLOCK_MONOTONIC,&start); //flight simulation start
    parent_process(pids ,pipe_fd, shm,start); // parent starts to give seats and confirm there booking
    cleanup(shm, shm_id);  //clean shared mem
    pclose(gnuplotPIPE);  //close the pipe to gnuplot 

    return 0;
}
